<?php

namespace Drupal\openai_batch;

use Drupal\ai\AiProviderPluginManager;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\Exception\FileException;
use Drupal\Core\File\FileExists;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\file\FileRepositoryInterface;
use Drupal\openai_batch\Entity\OpenAiBatchRequest;
use OpenAI\Client;
use OpenAI\Responses\Batches\BatchResponse;

final class OpenAiBatchService {

  public function getFiles(): array {
    $files_array = [];
    $response = $this->client->files()->list();
    foreach ($response->data as $result) {
      $file = $result->toArray();
      $files_array[] = $file;
    }
    return $files_array;
  }

  public function getInputFilesDirectory() {
    $main_dir = $this->config->get('system.file')->get('default_scheme') . "://openai_batch";
    $dir = $this->config->get('system.file')->get('default_scheme') . "://openai_batch/input_files";
    try {
      $this->fileSystem->prepareDirectory($main_dir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
      $this->fileSystem->prepareDirectory($dir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
    }
    catch (FileException $e) {
      // @TODO Log or set message or do something else.
    }
    return $dir;
  }

  public function getOutputFilesDirectory() {
    $main_dir = $this->config->get('system.file')->get('default_scheme') . "://openai_batch";
    $dir = $this->config->get('system.file')->get('default_scheme') . "://openai_batch/output_files";
    try {
      $this->fileSystem->prepareDirectory($main_dir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
      $this->fileSystem->prepareDirectory($dir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
    }
    catch (FileException $e) {
      // @TODO Log or set message or do something else.
    }
    return $dir;
  }

  public function saveJsonlFile($type, $file_id, $content) {
    $sub_dir = match ($type) {
      'input' => $this->getInputFilesDirectory(),
      'output' => $this->getOutputFilesDirectory(),
      default => throw new \Exception("Invalid batch file type: $type"),
    };
    $filepath = $sub_dir . "/$file_id.jsonl";
    $this->fileRepository->writeData($content, $filepath);
  }

  public function saveInputFile($file_id, $content) {
    $main_dir = $this->config->get('system.file')->get('default_scheme') . "://openai_batch";
    $input_dir = $this->config->get('system.file')->get('default_scheme') . "://openai_batch/input_files";
    try {
      $this->fileSystem->prepareDirectory($main_dir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
      $this->fileSystem->prepareDirectory($input_dir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
    }
    catch (FileException $e) {
      // @TODO Log or set message or do something else.
    }
    $filepath = $input_dir . "/$file_id.jsonl";
    $this->fileRepository->writeData($content, $filepath);
  }

  public function saveOutputFile($file_id, $content) {
    $main_dir = $this->config->get('system.file')->get('default_scheme') . "://openai_batch";
    $output_dir = $this->config->get('system.file')->get('default_scheme') . "://openai_batch/output_files";
    try {
      $this->fileSystem->prepareDirectory($main_dir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
      $this->fileSystem->prepareDirectory($output_dir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
    }
    catch (FileException $e) {
      // @TODO Log or set message or do something else.
    }
    $filepath = $output_dir . "/$file_id.jsonl";
    $this->fileRepository->writeData($content, $filepath, FileExists::Replace);
  }

  public function refreshBatchesStatusFormSubmitHandler(array &$form, FormStateInterface $form_state) {
    $batchBuilder = $this->getRefreshBatchesBatchBuilder();
    batch_set($batchBuilder->toArray());
  }

  public function getRefreshBatchesBatchBuilder(): BatchBuilder {
    $batchBuilder = new BatchBuilder();
    $batchBuilder->addOperation('\Drupal\openai_batch\Batch\OpenAiBatchOperations::fetchRemoteBatchesOperation', [
      $fetch_all = FALSE,
    ]);
    $batchBuilder->addOperation('\Drupal\openai_batch\Batch\OpenAiBatchOperations::updateLocalEntitiesOperation', [
      $create_missing = FALSE,
    ]);
    $batchBuilder
      ->setTitle($this->t('Fetching batches'))
      ->setFinishCallback('\Drupal\openai_batch\Batch\OpenAiBatchOperations::refreshBatchesOperationFinished')
      ->setErrorMessage(t('Batch has encountered an error'));
    return $batchBuilder;
  }

  public function updateOpenAiBatchRequestEntity(BatchResponse $batch_response, OpenAiBatchRequest $batch_entity, bool $save = TRUE): ? OpenAiBatchRequest {
    $remote_values = [
      'status' => (string) $batch_response->status,
      'output_file_id' => (string) $batch_response->outputFileId,
      'in_progress_at' => (string) $batch_response->inProgressAt,
      'expires_at' => (string) $batch_response->expiresAt,
      'finalizing_at' => (string) $batch_response->finalizingAt,
      'completed_at' => (string) $batch_response->completedAt,
      'failed_at' => (string) $batch_response->failedAt,
      'expired_at' => (string) $batch_response->expiredAt,
      'cancelling_at' => (string) $batch_response->cancellingAt,
      'cancelled_at' => (string) $batch_response->cancelledAt,
    ];
    $local_values = [
      'status' => (string) $batch_entity->get('status')->value,
      'output_file_id' => (string) $batch_entity->get('output_file_id')->value,
      'in_progress_at' => (string) $batch_entity->get('in_progress_at')->value,
      'expires_at' => (string) $batch_entity->get('expires_at')->value,
      'finalizing_at' => (string) $batch_entity->get('finalizing_at')->value,
      'completed_at' => (string) $batch_entity->get('completed_at')->value,
      'failed_at' => (string) $batch_entity->get('failed_at')->value,
      'expired_at' => (string) $batch_entity->get('expired_at')->value,
      'cancelling_at' => (string) $batch_entity->get('cancelling_at')->value,
      'cancelled_at' => (string) $batch_entity->get('cancelled_at')->value,
    ];
    if (($local_values !== $remote_values)) {
      $storage = $this->entityTypeManager->getStorage('openai_batch_request');
      $batch_entity = $storage->createRevision($batch_entity);
      $batch_entity->set('status', $batch_response->status);
      $batch_entity->set('output_file_id', $batch_response->outputFileId);
      $batch_entity->set('in_progress_at', $batch_response->inProgressAt);
      $batch_entity->set('expires_at', $batch_response->expiresAt);
      $batch_entity->set('finalizing_at', $batch_response->finalizingAt);
      $batch_entity->set('completed_at', $batch_response->completedAt);
      $batch_entity->set('failed_at', $batch_response->failedAt);
      $batch_entity->set('expired_at', $batch_response->expiredAt);
      $batch_entity->set('cancelling_at', $batch_response->cancellingAt);
      $batch_entity->set('cancelled_at', $batch_response->cancelledAt);
      if ($save) {
        $batch_entity->save();
      }
      return $batch_entity;
    }
    return NULL;
  }

  public function createJsonlContent(array $data_array) {
    // Initialize an empty string to hold all JSON lines
    $jsonl_content = '';

    // Loop through each item in the data array
    foreach ($data_array as $data) {
      // Convert each data array to a JSON string and add a newline
      $jsonl_content .= json_encode($data, JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE) . PHP_EOL;
    }
    return $jsonl_content;
  }

  private Client $client;

  private ConfigFactoryInterface $config;

  private FileSystemInterface $fileSystem;

  private FileRepositoryInterface $fileRepository;

  private EntityTypeManagerInterface $entityTypeManager;

  private StateInterface $state;

  public function __construct(AiProviderPluginManager $aiProvider, ConfigFactoryInterface $config, FileSystemInterface $file_system, FileRepositoryInterface $file_repository, EntityTypeManagerInterface $entity_type_manager, StateInterface $state) {
    /** @var \Drupal\ai\Plugin\ProviderProxy $provider */
    $provider = $aiProvider->createInstance("openai");
    $this->client = $provider->getClient();
    $this->config = $config;
    $this->fileSystem = $file_system;
    $this->fileRepository = $file_repository;
    $this->entityTypeManager = $entity_type_manager;
    $this->state = $state;
  }

  public function getClient(): Client {
    return $this->client;
  }

  use StringTranslationTrait;

}
