<?php

namespace Drupal\openai_batch\Plugin\Action;

use Drupal\ai\AiProviderPluginManager;
use Drupal\Core\Action\Attribute\Action;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TempStore\PrivateTempStore;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\openai_batch\DTO\OpenAiBatchRequestItemConfig;
use Drupal\openai_batch\Entity\OpenAiBatchRequest;
use Drupal\openai_batch\OpenAiBatchProcessorPluginManager;
use Drupal\openai_batch\OpenAiBatchService;
use Drupal\openai_batch\Plugin\Derivative\OpenAiBatchVBOActionDerivative;
use Drupal\views_bulk_operations\Action\ViewsBulkOperationsActionBase;
use Drupal\views_bulk_operations\Action\ViewsBulkOperationsPreconfigurationInterface;
use OpenAI\Client;
use Symfony\Component\DependencyInjection\ContainerInterface;

#[Action(
  id: 'openai_batch_vbo_action',
  action_label: new TranslatableMarkup('OpenAI Batch VBO Action'),
  deriver: OpenAiBatchVBOActionDerivative::class
)]
class OpenAiBatchVBOAction extends ViewsBulkOperationsActionBase implements ContainerFactoryPluginInterface, PluginFormInterface, ViewsBulkOperationsPreconfigurationInterface {

  // DONE abstracting out.
  private function getRowsFromView() {
    $this->view->render($this->view->current_display);
    $index = $this->context['sandbox']['processed'];
    $rows = [];
    $request_item_config = new OpenAiBatchRequestItemConfig(
      model: $this->configuration['model'],
      maxTokens: $this->configuration['configuration_max_tokens'],
      temperature: $this->configuration['configuration_temperature'],
      frequencyPenalty: $this->configuration['configuration_frequency_penalty'],
      presencePenalty: $this->configuration['configuration_presence_penalty'],
      topP: $this->configuration['configuration_top_p'],
      systemPrompt: $this->configuration['system_prompt'],
      userPrompt: $this->configuration['user_prompt'],
    );
    foreach ($this->view->result as $row) {
      $entity = $row->_entity;
      $batch_command = $this->openAiBatchProcessor->getOpenAiBatchCommandForEntity(
        entity_type: $entity->getEntityTypeId(),
        entity_id: $entity->id(),
        request_item_config: $request_item_config,
      );
      $rows[$index]['batch_command'] = $batch_command;
      $index++;
    }
    return $rows;
  }

  // DONE abstracting out.
  public function executeMultiple(array $objects) {
    unset($objects);
    if (empty($this->view->result)) {
      return;
    }

    $rows = $this->getRowsFromView();
    $processed = $this->context['sandbox']['processed'] + count($rows);
    $this->saveRowsToStore($rows);

    // Send the batch, if the last row has been processed.
    if (!isset($this->context['sandbox']['total']) || $processed >= $this->context['sandbox']['total']) {
      $this->sendBatch();
    }
  }

  // DONE abstracting out.
  private function saveRowsToStore(array $rows) {
    $this->tempStore->set($this->getCid(), $rows);
    unset($rows);
  }

  // DONE abstracting out.
  private function getCid() {
    if (!isset($this->context['sandbox']['cid_prefix'])) {
      $cid = $this->createCid();
      $this->context['sandbox']['cid_prefix'] = $cid;
    }
    return $this->context['sandbox']['cid_prefix'] . $this->context['sandbox']['current_batch'];
  }

  // DONE abstracting out.
  private function createCid(): string {
    $view_id = $this->context['view_id'];
    $display_id = $this->context['display_id'];
    $action_id = $this->context['action_id'];
    $cid = "$view_id:$display_id:$action_id:'" . json_encode(array_keys($this->context['list']));
    $time = microtime(TRUE);
    $cid = md5($cid . ":$time") . ':';
    return $cid;
  }

  // DONE abstracting out.
  protected function getCurrentRows() {
    $rows = [];
    for ($i = 1; $i <= $this->context['sandbox']['current_batch']; $i++) {
      $chunk = $this->tempStore->get($this->context['sandbox']['cid_prefix'] . $i);
      if ($chunk) {
        $rows = array_merge($rows, $chunk);
        $this->tempStore->delete($this->context['sandbox']['cid_prefix'] . $i);
      }
    }
    return $rows;
  }

  // DONE abstracting out.
  public function buildPreConfigurationForm(array $element, array $values, FormStateInterface $form_state): array {
    return $this->openAiBatchProcessor->buildPreConfigurationForm($element, $values, $form_state);
  }

  // DONE abstracting out.
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $llm_instance = $this->aiProviderManager->createInstance("openai");
    return $this->openAiBatchProcessor->buildConfigurationForm($form, $form_state, $this->context, $llm_instance);
  }

  // DONE abstracting out.
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->openAiBatchProcessor->validateConfigurationForm($form, $form_state);
  }

  protected function sendBatch() {
    $rows = $this->getCurrentRows();
    $batch = [];
    foreach ($rows as $row) {
      $batch_command = $row["batch_command"];
      $batch[] = $batch_command;
    }

    $tmp_filename = uniqid('openai_batch_', TRUE);
    $tmp_filepath = sys_get_temp_dir() . '/' . $tmp_filename . '.jsonl';
    $jsonl_content = $this->openAiBatchService->createJsonlContent($batch);
    file_put_contents($tmp_filepath, $jsonl_content);

    $upload_response = $this->client->files()->upload([
      'purpose' => 'batch',
      'file' => fopen($tmp_filepath, 'r'),
    ]);
    $file_id = $upload_response->id;

    $batch_response = $this->client->batches()->create(parameters: [
      'input_file_id' => $file_id,
      'endpoint' => '/v1/chat/completions',
      'completion_window' => '24h',
      'metadata' => [
        'plugin_id' => $this->getPluginId(),
        'uid' => $this->currentUser->id(),
      ],
    ]
    );
    $this->openAiBatchService->saveJsonlFile('input', $file_id, $jsonl_content);
    unlink($tmp_filepath);

    $batch_response_id = $batch_response->id;
    $batch_request_entity = OpenAiBatchRequest::create([
      'id' => $batch_response_id,
      'items' => count($batch),
      'items_processed' => 0,
      'processor' => $this->openAiBatchProcessor->getPluginId(),
      // @TODO: make this configurable
      'autosync' => TRUE,
      'status' => $batch_response->status,
      'input_file_id' => $batch_response->inputFileId,
      'output_file_id' => $batch_response->outputFileId,
      'created_at' => $batch_response->createdAt,
      'expires_at' => $batch_response->expiresAt,
      'expired_at' => $batch_response->expiredAt,
      'endpoint' => $batch_response->endpoint,
      'metadata' => json_encode(['vbo_configuration' => $this->configuration]),
    ]);
    $batch_request_entity->save();

    $this->messenger->addMessage($this->t('Request sent to ChatGPT, please wait for the batch processing to finish...'));
  }

  protected PrivateTempStore $tempStore;

  protected EntityTypeManagerInterface $entityTypeManager;

  protected AiProviderPluginManager $aiProviderManager;

  protected EntityFieldManagerInterface $entityFieldManager;

  protected OpenAiBatchVBOActionInterface $openAiBatchProcessor;

  private Client $client;

  private OpenAiBatchService $openAiBatchService;

  private AccountProxyInterface $currentUser;

  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    PrivateTempStoreFactory $temp_store_factory,
    EntityTypeManagerInterface $entity_type_manager,
    MessengerInterface $messenger,
    EntityFieldManagerInterface $entity_field_manager,
    AiProviderPluginManager $ai_provider_plugin_manager,
    OpenAiBatchProcessorPluginManager $openai_batch_processor_plugin_manager,
    OpenAiBatchService $openai_batch_service,
    AccountProxyInterface $current_user,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->tempStore = $temp_store_factory->get('openai_batch_' . $this->getPluginId());
    $this->entityTypeManager = $entity_type_manager;
    $this->messenger = $messenger;
    $this->entityFieldManager = $entity_field_manager;
    $this->aiProviderManager = $ai_provider_plugin_manager;
    $openai_batch_processor_id = $plugin_definition["openai_batch_processor_id"];
    $this->openAiBatchProcessor = $openai_batch_processor_plugin_manager->createInstance($openai_batch_processor_id);
    $provider = $ai_provider_plugin_manager->createInstance("openai");
    $this->client = $provider->getClient();
    $this->openAiBatchService = $openai_batch_service;
    $this->openAiBatchProcessor->setVboConfiguration($this->configuration);
    $this->currentUser = $current_user;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('tempstore.private'),
      $container->get('entity_type.manager'),
      $container->get('messenger'),
      $container->get('entity_field.manager'),
      $container->get('ai.provider'),
      $container->get('plugin.manager.openai_batch_processor'),
      $container->get('openai_batch.batch_service'),
      $container->get('current_user'),
    );
  }

  public function access($object, ?AccountInterface $account = NULL, $return_as_object = FALSE) {
    return $object->access('update', $account, $return_as_object);
  }

  public function execute($entity = NULL) {
    $this->executeMultiple([$entity]);
  }

}
